(in-package #:org.shirakumo.fraf.steamworks)

(defclass steammusicremote (c-managed-object interface)
  ((remote-handle :accessor remote-handle))
  (:default-initargs :free-on-gc T))

(defmethod initialize-instance :after ((interface steammusicremote) &key (name "remote"))
  (steam::music-remote-register-steam-music-remote (handle interface) name))

(defmethod allocate-handle ((interface steammusicremote) &key (version T) steamworks)
  (get-interface-handle* steamworks 'steam::client-get-isteam-music-remote
                         (t-or version STEAM::STEAMMUSICREMOTE-INTERFACE-VERSION)))

(defmethod free-handle-function ((interface steammusicremote) handle)
  (lambda () (steam::music-remote-deregister-steam-music-remote handle)))

(define-interface-method steammusicremote (setf activated) (value steam::music-remote-bactivation-success)
  (check-invalid NIL result 'steam::music-remote-bactivation-success))
(define-interface-method steammusicremote remote-p (steam::music-remote-bis-current-music-remote))
(define-interface-method steammusicremote entry-changed-p (steam::music-remote-current-entry-did-change))
(define-interface-method steammusicremote (setf entry-available-p) (value steam::music-remote-current-entry-is-available)
  (check-invalid NIL result 'steam::music-remote-current-entry-is-available))
(define-interface-method steammusicremote entry-will-change-p (steam::music-remote-current-entry-will-change))
(define-interface-method steammusicremote (setf looped-p) (value steam::music-remote-enable-looped))
(define-interface-method steammusicremote (setf playlists-enabled-p) (value steam::music-remote-enable-playlists)
  (check-invalid NIL result 'steam::music-remote-enable-playlists))
(define-interface-method steammusicremote (setf play-next-enabled-p) (value steam::music-remote-enable-play-next)
  (check-invalid NIL result 'steam::music-remote-enable-play-next))
(define-interface-method steammusicremote (setf play-previous-enabled-p) (value steam::music-remote-enable-play-previous)
  (check-invalid NIL result 'steam::music-remote-enable-play-previous))
(define-interface-method steammusicremote (setf queue-enabled-p) (value steam::music-remote-enable-queue)
  (check-invalid NIL result 'steam::music-remote-enable-queue))
(define-interface-method steammusicremote (setf shuffle-enabled-p) (value steam::music-remote-enable-shuffled)
  (check-invalid NIL result 'steam::music-remote-enable-shuffled))
(define-interface-method steammusicremote playlist-changed-p (steam::music-remote-playlist-did-change))
(define-interface-method steammusicremote playlist-will-change-p (steam::music-remote-playlist-will-change))
(define-interface-method steammusicremote queue-changed-p (steam::music-remote-queue-did-change))
(define-interface-method steammusicremote queue-will-change-p (steam::music-remote-queue-will-change))
(define-interface-method steammusicremote reset-playlist (steam::music-remote-reset-playlist-entries)
  (check-invalid NIL result 'steam::music-remote-reset-playlist-entries))
(define-interface-method steammusicremote reset-queue (steam::music-remote-reset-queue-entries)
  (check-invalid NIL result 'steam::music-remote-reset-queue-entries))
(define-interface-method steammusicremote (setf current-playlist-entry) ((value integer) steam::music-remote-set-current-playlist-entry)
  (check-invalid NIL result 'steam::music-remote-set-current-playlist-entry))
(define-interface-method steammusicremote (setf current-queue-entry) ((value integer) steam::music-remote-set-current-queue-entry)
  (check-invalid NIL result 'steam::music-remote-set-current-queue-entry))
(define-interface-method steammusicremote (setf display-name) ((value string) steam::music-remote-set-display-name)
  (check-invalid NIL result 'steam::music-remote-set-display-name))
(define-interface-method steammusicremote (setf elapsed-seconds) ((value integer) steam::music-remote-update-current-entry-elapsed-seconds)
  (check-invalid NIL result 'steam::music-remote-update-current-entry-elapsed-seconds))
(define-interface-method steammusicremote (setf entry-text) ((value string) steam::music-remote-update-current-entry-text)
  (check-invalid NIL result 'steam::music-remote-update-current-entry-text))
(define-interface-method steammusicremote (setf looped-p) (value steam::music-remote-update-looped)
  (check-invalid NIL result 'steam::music-remote-update-looped))
(define-interface-method steammusicremote (setf playback-status) (value steam::music-remote-update-playback-status)
  (check-invalid NIL result 'steam::music-remote-update-playback-status))
(define-interface-method steammusicremote (setf shuffled-p) (value steam::music-remote-update-shuffled)
  (check-invalid NIL result 'steam::music-remote-update-shuffled))
(define-interface-method steammusicremote (setf volume) ((value float) steam::music-remote-update-volume)
  (check-invalid NIL result 'steam::music-remote-update-volume))
(define-interface-method steammusicremote (setf queue-entry) ((value string) steam::music-remote-set-queue-entry (id integer) (position integer))
  (check-invalid NIL result 'steam::music-remote-set-queue-entry))
(define-interface-method steammusicremote (setf playlist-entry) ((value string) steam::music-remote-set-playlist-entry (id integer) (position integer))
  (check-invalid NIL result 'steam::music-remote-set-playlist-entry))

(defmethod (setf icon) ((image pathname) (remote steammusicremote))
  (setf (icon remote) (read-file-to-sharable-byte-vector image)))

(defmethod (setf icon) ((image vector) (remote steammusicremote))
  (cffi:with-pointer-to-vector-data (buffer image)
    (steam::music-remote-set-pngicon-64x64 (handle remote) buffer (length image))))

(defmethod (setf cover-art) ((image pathname) (remote steammusicremote))
  (setf (cover-art remote) (read-file-to-sharable-byte-vector image)))

(defmethod (setf cover-art) ((image vector) (remote steammusicremote))
  (cffi:with-pointer-to-vector-data (buffer image)
    (steam::music-remote-update-current-entry-cover-art (handle remote) buffer (length image))))
